"""
    :codeauthor: jmoney <justin@saltstack.com>
"""


import pytest

import salt.channel.client
import salt.modules.cp as cp
import salt.utils.files
import salt.utils.platform
import salt.utils.templates as templates
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, Mock, mock_open, patch


@pytest.fixture()
def configure_loader_modules():
    return {cp: {"__opts__": {"saltenv": None}}}


def test__render_filenames_undefined_template():
    """
    Test if _render_filenames fails upon getting a template not in
    TEMPLATE_REGISTRY.
    """
    path = "/srv/salt/saltines"
    dest = "/srv/salt/cheese"
    saltenv = "base"
    template = "biscuits"
    ret = (path, dest)
    pytest.raises(
        CommandExecutionError, cp._render_filenames, path, dest, saltenv, template
    )


def test__render_filenames_render_failed():
    """
    Test if _render_filenames fails when template rendering fails.
    """
    path = "salt://saltines"
    dest = "/srv/salt/cheese"
    saltenv = "base"
    template = "jinja"
    file_data = "Remember to keep your files well salted."
    mock_jinja = lambda *args, **kwargs: {"result": False, "data": file_data}
    with patch.dict(templates.TEMPLATE_REGISTRY, {"jinja": mock_jinja}):
        with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
            pytest.raises(
                CommandExecutionError,
                cp._render_filenames,
                path,
                dest,
                saltenv,
                template,
            )


def test__render_filenames_success():
    """
    Test if _render_filenames succeeds.
    """
    path = "salt://saltines"
    dest = "/srv/salt/cheese"
    saltenv = "base"
    template = "jinja"
    file_data = "/srv/salt/biscuits"
    mock_jinja = lambda *args, **kwargs: {"result": True, "data": file_data}
    ret = (file_data, file_data)  # salt.utils.files.fopen can only be mocked once
    with patch.dict(templates.TEMPLATE_REGISTRY, {"jinja": mock_jinja}):
        with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
            assert cp._render_filenames(path, dest, saltenv, template) == ret
            # saltenv=None should default to "base" when not set in config
            assert (
                cp._render_filenames(path, dest, saltenv=None, template=template) == ret
            )


def test_get_file_not_found():
    """
    Test if get_file can't find the file.
    """
    with patch("salt.modules.cp.hash_file", MagicMock(return_value=False)):
        path = "salt://saltines"
        dest = "/srv/salt/cheese"
        ret = ""
        assert cp.get_file(path, dest) == ret


def test_get_file_str_success():
    """
    Test if get_file_str succeeds.
    """
    path = "salt://saltines"
    dest = "/srv/salt/cheese/saltines"
    file_data = "Remember to keep your files well salted."
    saltenv = "base"
    ret = file_data
    with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        with patch("salt.modules.cp.cache_file", MagicMock(return_value=dest)):
            assert cp.get_file_str(path, dest) == ret


def test_push_non_absolute_path():
    """
    Test if push fails on a non absolute path.
    """
    path = "../saltines"
    ret = False

    assert cp.push(path) == ret


def test_push_dir_non_absolute_path():
    """
    Test if push_dir fails on a non absolute path.
    """
    path = "../saltines"
    ret = False

    assert cp.push_dir(path) == ret


def test_push():
    """
    Test if push works with good posix path.
    """
    filename = "/saltines/test.file"
    if salt.utils.platform.is_windows():
        filename = "C:\\saltines\\test.file"
    with patch(
        "salt.modules.cp.os.path",
        MagicMock(isfile=Mock(return_value=True), wraps=cp.os.path),
    ), patch(
        "salt.modules.cp.os.path",
        MagicMock(getsize=MagicMock(return_value=10), wraps=cp.os.path),
    ), patch.multiple(
        "salt.modules.cp",
        _auth=MagicMock(**{"return_value.gen_token.return_value": "token"}),
        __opts__={"id": "abc", "file_buffer_size": 10},
    ), patch(
        "salt.utils.files.fopen", mock_open(read_data=b"content")
    ) as m_open, patch(
        "salt.channel.client.ReqChannel.factory", MagicMock()
    ) as req_channel_factory_mock:
        response = cp.push(filename)
        assert response, response
        num_opens = len(m_open.filehandles[filename])
        assert num_opens == 1, num_opens
        fh_ = m_open.filehandles[filename][0]
        assert fh_.read.call_count == 2, fh_.read.call_count
        req_channel_factory_mock().__enter__().send.assert_called_once_with(
            dict(
                loc=fh_.tell(),  # pylint: disable=resource-leakage
                cmd="_file_recv",
                tok="token",
                path=["saltines", "test.file"],
                size=10,
                data=b"",  # data is empty here because load['data'] is overwritten
                id="abc",
            )
        )



"""
Module for Sending Messages via SMTP

.. versionadded:: 2014.7.0

:depends:   - smtplib python module
:configuration: This module can be used by either passing a jid and password
    directly to send_message, or by specifying the name of a configuration
    profile in the minion config, minion pillar, or master config.

    For example:

    .. code-block:: yaml

        my-smtp-login:
            smtp.server: smtp.domain.com
            smtp.tls: True
            smtp.sender: admin@domain.com
            smtp.username: myuser
            smtp.password: verybadpass

    The resourcename refers to the resource that is using this account. It is
    user-definable, and optional. The following configurations are both valid:

    .. code-block:: yaml

        my-smtp-login:
            smtp.server: smtp.domain.com
            smtp.tls: True
            smtp.sender: admin@domain.com
            smtp.username: myuser
            smtp.password: verybadpass

        another-smtp-login:
            smtp.server: smtp.domain.com
            smtp.tls: True
            smtp.sender: admin@domain.com
            smtp.username: myuser
            smtp.password: verybadpass

"""


import logging
import os
import socket

import salt.utils.files

log = logging.getLogger(__name__)

HAS_LIBS = False
try:
    import email.mime.application
    import email.mime.multipart
    import email.mime.text
    import smtplib

    HAS_LIBS = True
except ImportError:
    pass


__virtualname__ = "smtp"


def __virtual__():
    """
    Only load this module if smtplib is available on this minion.
    """
    if HAS_LIBS:
        return __virtualname__
    return (False, "This module is only loaded if smtplib is available")


def send_msg(
    recipient,
    message,
    subject="Message from Salt",
    sender=None,
    server=None,
    use_ssl="True",
    username=None,
    password=None,
    profile=None,
    attachments=None,
):
    """
    Send a message to an SMTP recipient. To send a message to multiple \
    recipients, the recipients should be in a comma-seperated Python string. \
    Designed for use in states.

    CLI Examples:

    .. code-block:: bash

        salt '*' smtp.send_msg 'admin@example.com' 'This is a salt module test' profile='my-smtp-account'
        salt '*' smtp.send_msg 'admin@example.com,admin2@example.com' 'This is a salt module test for multiple recipients' profile='my-smtp-account'
        salt '*' smtp.send_msg 'admin@example.com' 'This is a salt module test' username='myuser' password='verybadpass' sender='admin@example.com' server='smtp.domain.com'
        salt '*' smtp.send_msg 'admin@example.com' 'This is a salt module test' username='myuser' password='verybadpass' sender='admin@example.com' server='smtp.domain.com' attachments="['/var/log/messages']"
    """
    if profile:
        creds = __salt__["config.option"](profile)
        server = creds.get("smtp.server")
        use_ssl = creds.get("smtp.tls")
        sender = creds.get("smtp.sender")
        username = creds.get("smtp.username")
        password = creds.get("smtp.password")

    if attachments:
        msg = email.mime.multipart.MIMEMultipart()
        msg.attach(email.mime.text.MIMEText(message))
    else:
        msg = email.mime.text.MIMEText(message)
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = recipient
    recipients = [r.strip() for r in recipient.split(",")]

    try:
        if use_ssl in ["True", "true"]:
            smtpconn = smtplib.SMTP_SSL(server)
        else:
            smtpconn = smtplib.SMTP(server)

    except socket.gaierror as _error:
        log.debug("Exception: %s", _error)
        return False

    if use_ssl not in ("True", "true"):
        smtpconn.ehlo()
        if smtpconn.has_extn("STARTTLS"):
            try:
                smtpconn.starttls()
            except smtplib.SMTPHeloError:
                log.debug("The server didn’t reply properly to the HELO greeting.")
                return False
            except smtplib.SMTPException:
                log.debug("The server does not support the STARTTLS extension.")
                return False
            except RuntimeError:
                log.debug(
                    "SSL/TLS support is not available to your Python interpreter."
                )
                return False
            smtpconn.ehlo()

    if username and password:
        try:
            smtpconn.login(username, password)
        except smtplib.SMTPAuthenticationError as _error:
            log.debug("SMTP Authentication Failure")
            return False

    if attachments:
        for f in attachments:
            name = os.path.basename(f)
            with salt.utils.files.fopen(f, "rb") as fin:
                att = email.mime.application.MIMEApplication(fin.read(), Name=name)
            att["Content-Disposition"] = 'attachment; filename="{}"'.format(name)
            msg.attach(att)

    try:
        smtpconn.sendmail(sender, recipients, msg.as_string())
    except smtplib.SMTPRecipientsRefused:
        log.debug("All recipients were refused.")
        return False
    except smtplib.SMTPHeloError:
        log.debug("The server didn’t reply properly to the HELO greeting.")
        return False
    except smtplib.SMTPSenderRefused:
        log.debug("The server didn’t accept the %s.", sender)
        return False
    except smtplib.SMTPDataError:
        log.debug("The server replied with an unexpected error code.")
        return False

    smtpconn.quit()
    return True



"""
    :codeauthor: :email:`Roald Nefs (info@roaldnefs.com)`

    Test cases for salt.modules.telegram.
"""


import pytest

import salt.modules.telegram as telegram
from tests.support.mock import MagicMock, Mock


class RequestMock(Mock):
    """
    Request Mock
    """

    def get(self, *args, **kwargs):
        return RequestResponseMock()

    def put(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return RequestPutResponseMock()

    def delete(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return RequestResponseMock()


class RequestResponseMock(Mock):
    """
    Request Response Mock
    """

    def json(self):
        return [
            {"url": "http://example.org", "_id": 1234},
        ]


class RequestPutResponseMock(Mock):
    """
    Request Put Response Mock
    """

    ok = True

    def json(self):
        return {"_id": 4321}


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "__salt__": {
            "config.get": MagicMock(
                return_value={
                    "telegram": {
                        "chat_id": "123456789",
                        "token": "000000000:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                    }
                }
            ),
            "requests.put": Mock(),
        },
        "requests": RequestMock(),
    }
    return {telegram: module_globals}


def test_post_message():
    """
    Test the post_message function.
    """
    message = "Hello World!"
    assert telegram.post_message(message)



"""
    salt.syspaths
    ~~~~~~~~~~~~~

    Salt's defaults system paths

    This module allows defining Salt's default paths at build time by writing a
    ``_syspath.py`` file to the filesystem. This is useful, for example, for
    setting platform-specific defaults that differ from the standard Linux
    paths.

    These values are static values and must be considered as secondary to any
    paths that are set in the master/minion config files.
"""


import logging
import os
import os.path
import sys

import salt.utils.platform
import salt.utils.win_reg

if salt.utils.platform.is_junos():
    __PLATFORM = "junos"
else:
    __PLATFORM = sys.platform.lower()

typo_warning = True
log = logging.getLogger(__name__)
EXPECTED_VARIABLES = (
    "ROOT_DIR",
    "CONFIG_DIR",
    "CACHE_DIR",
    "SOCK_DIR",
    "SRV_ROOT_DIR",
    "BASE_FILE_ROOTS_DIR",
    "HOME_DIR",
    "BASE_PILLAR_ROOTS_DIR",
    "BASE_THORIUM_ROOTS_DIR",
    "BASE_MASTER_ROOTS_DIR",
    "LIB_STATE_DIR",
    "LOGS_DIR",
    "PIDFILE_DIR",
    "SPM_PARENT_PATH",
    "SPM_FORMULA_PATH",
    "SPM_PILLAR_PATH",
    "SPM_REACTOR_PATH",
    "SHARE_DIR",
)

try:
    # Let's try loading the system paths from the generated module at
    # installation time.
    import salt._syspaths as __generated_syspaths  # pylint: disable=no-name-in-module
except ImportError:
    import types

    __generated_syspaths = types.ModuleType("salt._syspaths")
    for key in EXPECTED_VARIABLES:
        setattr(__generated_syspaths, key, None)
else:
    for key in EXPECTED_VARIABLES:
        if hasattr(__generated_syspaths, key):
            continue
        else:
            if typo_warning:
                log.warning("Possible Typo?")
                log.warning(
                    "To dissolve this warning add `[variable] = None` to _syspaths.py"
                )
            typo_warning = False
            log.warning("Variable %s is missing, value set to None", key)
            setattr(
                __generated_syspaths, key, None
            )  # missing variables defaulted to None

# Let's find out the path of this module
if "SETUP_DIRNAME" in globals():
    # This is from the exec() call in Salt's setup.py
    # pylint: disable=undefined-variable
    __THIS_FILE = os.path.join(SETUP_DIRNAME, "salt", "syspaths.py")
    # pylint: enable=undefined-variable
else:
    __THIS_FILE = __file__


# These values are always relative to salt's installation directory
INSTALL_DIR = os.path.dirname(os.path.realpath(__THIS_FILE))
CLOUD_DIR = os.path.join(INSTALL_DIR, "cloud")
BOOTSTRAP = os.path.join(CLOUD_DIR, "deploy", "bootstrap-salt.sh")


def _get_windows_root_dir():
    # Try to get the root directory location from the registry
    # This key will be created by the NullSoft installer
    # If salt is currently installed in C:\salt and the user performs an
    # upgrade, then this key will be set to C:\salt
    root_dir = salt.utils.win_reg.read_value(
        hive="HKLM", key="SOFTWARE\\Salt Project\\salt", vname="root_dir"
    )
    if root_dir["success"]:
        # Make sure vdata contains something
        if root_dir["vdata"]:
            return os.path.expandvars(root_dir["vdata"])

    # If this key does not exist, then salt was not installed using the
    # new method installer. Could be pip or setup.py or an older version of the
    # installer.
    log.debug("Failed to get ROOT_DIR from registry. %s", root_dir["comment"])
    # Check for C:\salt\conf
    old_root = "\\".join([os.environ["SystemDrive"], "salt", "conf"])
    dflt_root = os.path.join(os.environ["ProgramData"], "Salt Project", "Salt")
    if os.path.isdir(old_root):
        # If the old config location is present use it
        log.debug("ROOT_DIR: %s", os.path.dirname(old_root))
        return os.path.dirname(old_root)
    else:
        # If not, then default to ProgramData
        log.debug("ROOT_DIR: %s", dflt_root)
        return dflt_root


ROOT_DIR = __generated_syspaths.ROOT_DIR
if ROOT_DIR is None:
    # The installation time value was not provided, let's define the default
    if __PLATFORM.startswith("win"):
        ROOT_DIR = _get_windows_root_dir()
    else:
        ROOT_DIR = "/"

CONFIG_DIR = __generated_syspaths.CONFIG_DIR
if CONFIG_DIR is None:
    if __PLATFORM.startswith("win"):
        CONFIG_DIR = os.path.join(ROOT_DIR, "conf")
    elif "freebsd" in __PLATFORM:
        CONFIG_DIR = os.path.join(ROOT_DIR, "usr", "local", "etc", "salt")
    elif "junos" in __PLATFORM:
        CONFIG_DIR = os.path.join(ROOT_DIR, "var", "local", "salt", "etc")
    elif "netbsd" in __PLATFORM:
        CONFIG_DIR = os.path.join(ROOT_DIR, "usr", "pkg", "etc", "salt")
    elif "sunos5" in __PLATFORM:
        CONFIG_DIR = os.path.join(ROOT_DIR, "opt", "local", "etc", "salt")
    else:
        CONFIG_DIR = os.path.join(ROOT_DIR, "etc", "salt")

SHARE_DIR = __generated_syspaths.SHARE_DIR
if SHARE_DIR is None:
    if __PLATFORM.startswith("win"):
        SHARE_DIR = os.path.join(ROOT_DIR, "share")
    elif "freebsd" in __PLATFORM:
        SHARE_DIR = os.path.join(ROOT_DIR, "usr", "local", "share", "salt")
    elif "junos" in __PLATFORM:
        SHARE_DIR = os.path.join(ROOT_DIR, "var", "local", "salt", "share")
    elif "netbsd" in __PLATFORM:
        SHARE_DIR = os.path.join(ROOT_DIR, "usr", "share", "salt")
    elif "sunos5" in __PLATFORM:
        SHARE_DIR = os.path.join(ROOT_DIR, "usr", "share", "salt")
    else:
        SHARE_DIR = os.path.join(ROOT_DIR, "usr", "share", "salt")

CACHE_DIR = __generated_syspaths.CACHE_DIR
if CACHE_DIR is None:
    if "junos" in __PLATFORM:
        CACHE_DIR = os.path.join(ROOT_DIR, "var", "local", "salt", "cache")
    else:
        CACHE_DIR = os.path.join(ROOT_DIR, "var", "cache", "salt")

SOCK_DIR = __generated_syspaths.SOCK_DIR
if SOCK_DIR is None:
    if "junos" in __PLATFORM:
        SOCK_DIR = os.path.join(ROOT_DIR, "var", "local", "salt", "run")
    else:
        SOCK_DIR = os.path.join(ROOT_DIR, "var", "run", "salt")

SRV_ROOT_DIR = __generated_syspaths.SRV_ROOT_DIR
if SRV_ROOT_DIR is None:
    SRV_ROOT_DIR = os.path.join(ROOT_DIR, "srv")

BASE_FILE_ROOTS_DIR = __generated_syspaths.BASE_FILE_ROOTS_DIR
if BASE_FILE_ROOTS_DIR is None:
    BASE_FILE_ROOTS_DIR = os.path.join(SRV_ROOT_DIR, "salt")

BASE_PILLAR_ROOTS_DIR = __generated_syspaths.BASE_PILLAR_ROOTS_DIR
if BASE_PILLAR_ROOTS_DIR is None:
    BASE_PILLAR_ROOTS_DIR = os.path.join(SRV_ROOT_DIR, "pillar")

BASE_THORIUM_ROOTS_DIR = __generated_syspaths.BASE_THORIUM_ROOTS_DIR
if BASE_THORIUM_ROOTS_DIR is None:
    BASE_THORIUM_ROOTS_DIR = os.path.join(SRV_ROOT_DIR, "thorium")

BASE_MASTER_ROOTS_DIR = __generated_syspaths.BASE_MASTER_ROOTS_DIR
if BASE_MASTER_ROOTS_DIR is None:
    BASE_MASTER_ROOTS_DIR = os.path.join(SRV_ROOT_DIR, "salt-master")

LIB_STATE_DIR = __generated_syspaths.LIB_STATE_DIR
if LIB_STATE_DIR is None:
    LIB_STATE_DIR = CONFIG_DIR

LOGS_DIR = __generated_syspaths.LOGS_DIR
if LOGS_DIR is None:
    LOGS_DIR = os.path.join(ROOT_DIR, "var", "log", "salt")

PIDFILE_DIR = __generated_syspaths.PIDFILE_DIR
if PIDFILE_DIR is None:
    if "junos" in __PLATFORM:
        PIDFILE_DIR = os.path.join(ROOT_DIR, "var", "local", "salt", "run")
    else:
        PIDFILE_DIR = os.path.join(ROOT_DIR, "var", "run")

SPM_PARENT_PATH = __generated_syspaths.SPM_PARENT_PATH
if SPM_PARENT_PATH is None:
    SPM_PARENT_PATH = os.path.join(SRV_ROOT_DIR, "spm")

SPM_FORMULA_PATH = __generated_syspaths.SPM_FORMULA_PATH
if SPM_FORMULA_PATH is None:
    SPM_FORMULA_PATH = os.path.join(SPM_PARENT_PATH, "salt")

SPM_PILLAR_PATH = __generated_syspaths.SPM_PILLAR_PATH
if SPM_PILLAR_PATH is None:
    SPM_PILLAR_PATH = os.path.join(SPM_PARENT_PATH, "pillar")

SPM_REACTOR_PATH = __generated_syspaths.SPM_REACTOR_PATH
if SPM_REACTOR_PATH is None:
    SPM_REACTOR_PATH = os.path.join(SPM_PARENT_PATH, "reactor")

HOME_DIR = __generated_syspaths.HOME_DIR
if HOME_DIR is None:
    HOME_DIR = os.path.expanduser("~")


__all__ = [
    "ROOT_DIR",
    "SHARE_DIR",
    "CONFIG_DIR",
    "CACHE_DIR",
    "SOCK_DIR",
    "SRV_ROOT_DIR",
    "BASE_FILE_ROOTS_DIR",
    "BASE_PILLAR_ROOTS_DIR",
    "BASE_MASTER_ROOTS_DIR",
    "BASE_THORIUM_ROOTS_DIR",
    "LIB_STATE_DIR",
    "LOGS_DIR",
    "PIDFILE_DIR",
    "INSTALL_DIR",
    "CLOUD_DIR",
    "BOOTSTRAP",
    "SPM_PARENT_PATH",
    "SPM_FORMULA_PATH",
    "SPM_PILLAR_PATH",
    "SPM_REACTOR_PATH",
]



"""
Azure (ARM) Utilities

.. versionadded:: 2019.2.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0rc6
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.4
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 0.30.0rc6
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 0.33.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 0.30.0rc6
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 0.30.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 0.30.0rc6
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.30.0rc6
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.32.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

"""

import importlib
import logging
import sys
from operator import itemgetter

import salt.config
import salt.loader
import salt.utils.args
import salt.utils.stringutils
import salt.utils.versions
import salt.version
from salt.exceptions import SaltInvocationError, SaltSystemExit

try:
    from azure.common.credentials import (
        ServicePrincipalCredentials,
        UserPassCredentials,
    )
    from msrestazure.azure_cloud import (
        MetadataEndpointError,
        get_cloud_from_metadata_endpoint,
    )

    HAS_AZURE = True
except ImportError:
    HAS_AZURE = False

__opts__ = salt.config.minion_config("/etc/salt/minion")
__salt__ = salt.loader.minion_mods(__opts__)

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_AZURE:
        return False
    else:
        return True


def _determine_auth(**kwargs):
    """
    Acquire Azure ARM Credentials
    """
    if "profile" in kwargs:
        azure_credentials = __salt__["config.option"](kwargs["profile"])
        kwargs.update(azure_credentials)

    service_principal_creds_kwargs = ["client_id", "secret", "tenant"]
    user_pass_creds_kwargs = ["username", "password"]

    try:
        if kwargs.get("cloud_environment") and kwargs.get(
            "cloud_environment"
        ).startswith("http"):
            cloud_env = get_cloud_from_metadata_endpoint(kwargs["cloud_environment"])
        else:
            cloud_env_module = importlib.import_module("msrestazure.azure_cloud")
            cloud_env = getattr(
                cloud_env_module, kwargs.get("cloud_environment", "AZURE_PUBLIC_CLOUD")
            )
    except (AttributeError, ImportError, MetadataEndpointError):
        raise sys.exit(
            "The Azure cloud environment {} is not available.".format(
                kwargs["cloud_environment"]
            )
        )

    if set(service_principal_creds_kwargs).issubset(kwargs):
        if not (kwargs["client_id"] and kwargs["secret"] and kwargs["tenant"]):
            raise SaltInvocationError(
                "The client_id, secret, and tenant parameters must all be "
                "populated if using service principals."
            )
        else:
            credentials = ServicePrincipalCredentials(
                kwargs["client_id"],
                kwargs["secret"],
                tenant=kwargs["tenant"],
                cloud_environment=cloud_env,
            )
    elif set(user_pass_creds_kwargs).issubset(kwargs):
        if not (kwargs["username"] and kwargs["password"]):
            raise SaltInvocationError(
                "The username and password parameters must both be "
                "populated if using username/password authentication."
            )
        else:
            credentials = UserPassCredentials(
                kwargs["username"], kwargs["password"], cloud_environment=cloud_env
            )
    elif "subscription_id" in kwargs:
        try:
            from msrestazure.azure_active_directory import MSIAuthentication

            credentials = MSIAuthentication(cloud_environment=cloud_env)
        except ImportError:
            raise SaltSystemExit(
                msg=(
                    "MSI authentication support not availabe (requires msrestazure >="
                    " 0.4.14)"
                )
            )

    else:
        raise SaltInvocationError(
            "Unable to determine credentials. "
            "A subscription_id with username and password, "
            "or client_id, secret, and tenant or a profile with the "
            "required parameters populated"
        )

    if "subscription_id" not in kwargs:
        raise SaltInvocationError("A subscription_id must be specified")

    subscription_id = salt.utils.stringutils.to_str(kwargs["subscription_id"])

    return credentials, subscription_id, cloud_env


def get_client(client_type, **kwargs):
    """
    Dynamically load the selected client and return a management client object
    """
    client_map = {
        "compute": "ComputeManagement",
        "authorization": "AuthorizationManagement",
        "dns": "DnsManagement",
        "storage": "StorageManagement",
        "managementlock": "ManagementLock",
        "monitor": "MonitorManagement",
        "network": "NetworkManagement",
        "policy": "Policy",
        "resource": "ResourceManagement",
        "subscription": "Subscription",
        "web": "WebSiteManagement",
    }

    if client_type not in client_map:
        raise SaltSystemExit(
            msg="The Azure ARM client_type {} specified can not be found.".format(
                client_type
            )
        )

    map_value = client_map[client_type]

    if client_type in ["policy", "subscription"]:
        module_name = "resource"
    elif client_type in ["managementlock"]:
        module_name = "resource.locks"
    else:
        module_name = client_type

    try:
        client_module = importlib.import_module("azure.mgmt." + module_name)
        # pylint: disable=invalid-name
        Client = getattr(client_module, "{}Client".format(map_value))
    except ImportError:
        raise sys.exit("The azure {} client is not available.".format(client_type))

    credentials, subscription_id, cloud_env = _determine_auth(**kwargs)

    if client_type == "subscription":
        client = Client(
            credentials=credentials,
            base_url=cloud_env.endpoints.resource_manager,
        )
    else:
        client = Client(
            credentials=credentials,
            subscription_id=subscription_id,
            base_url=cloud_env.endpoints.resource_manager,
        )

    client.config.add_user_agent("Salt/{}".format(salt.version.__version__))

    return client


def log_cloud_error(client, message, **kwargs):
    """
    Log an azurearm cloud error exception
    """
    try:
        cloud_logger = getattr(log, kwargs.get("azurearm_log_level"))
    except (AttributeError, TypeError):
        cloud_logger = getattr(log, "error")

    cloud_logger(
        "An AzureARM %s CloudError has occurred: %s", client.capitalize(), message
    )

    return


def paged_object_to_list(paged_object):
    """
    Extract all pages within a paged object as a list of dictionaries
    """
    paged_return = []
    while True:
        try:
            page = next(paged_object)
            paged_return.append(page.as_dict())
        except StopIteration:
            break

    return paged_return


def create_object_model(module_name, object_name, **kwargs):
    """
    Assemble an object from incoming parameters.
    """
    object_kwargs = {}

    try:
        model_module = importlib.import_module(
            "azure.mgmt.{}.models".format(module_name)
        )
        # pylint: disable=invalid-name
        Model = getattr(model_module, object_name)
    except ImportError:
        raise sys.exit(
            "The {} model in the {} Azure module is not available.".format(
                object_name, module_name
            )
        )

    if "_attribute_map" in dir(Model):
        for attr, items in Model._attribute_map.items():
            param = kwargs.get(attr)
            if param is not None:
                if items["type"][0].isupper() and isinstance(param, dict):
                    object_kwargs[attr] = create_object_model(
                        module_name, items["type"], **param
                    )
                elif items["type"][0] == "{" and isinstance(param, dict):
                    object_kwargs[attr] = param
                elif items["type"][0] == "[" and isinstance(param, list):
                    obj_list = []
                    for list_item in param:
                        if items["type"][1].isupper() and isinstance(list_item, dict):
                            obj_list.append(
                                create_object_model(
                                    module_name,
                                    items["type"][
                                        items["type"].index("[")
                                        + 1 : items["type"].rindex("]")
                                    ],
                                    **list_item
                                )
                            )
                        elif items["type"][1] == "{" and isinstance(list_item, dict):
                            obj_list.append(list_item)
                        elif not items["type"][1].isupper() and items["type"][1] != "{":
                            obj_list.append(list_item)
                    object_kwargs[attr] = obj_list
                else:
                    object_kwargs[attr] = param

    # wrap calls to this function to catch TypeError exceptions
    return Model(**object_kwargs)


def compare_list_of_dicts(old, new, convert_id_to_name=None):
    """
    Compare lists of dictionaries representing Azure objects. Only keys found in the "new" dictionaries are compared to
    the "old" dictionaries, since getting Azure objects from the API returns some read-only data which should not be
    used in the comparison. A list of parameter names can be passed in order to compare a bare object name to a full
    Azure ID path for brevity. If string types are found in values, comparison is case insensitive. Return comment
    should be used to trigger exit from the calling function.
    """
    ret = {}

    if not convert_id_to_name:
        convert_id_to_name = []

    if not isinstance(new, list):
        ret["comment"] = "must be provided as a list of dictionaries!"
        return ret

    if len(new) != len(old):
        ret["changes"] = {"old": old, "new": new}
        return ret

    try:
        local_configs, remote_configs = (
            sorted(config, key=itemgetter("name")) for config in (new, old)
        )
    except TypeError:
        ret["comment"] = "configurations must be provided as a list of dictionaries!"
        return ret
    except KeyError:
        ret["comment"] = 'configuration dictionaries must contain the "name" key!'
        return ret

    for idx, val in enumerate(local_configs):
        for key in val:
            local_val = val[key]
            if key in convert_id_to_name:
                remote_val = (
                    remote_configs[idx].get(key, {}).get("id", "").split("/")[-1]
                )
            else:
                remote_val = remote_configs[idx].get(key)
                if isinstance(local_val, str):
                    local_val = local_val.lower()
                if isinstance(remote_val, str):
                    remote_val = remote_val.lower()
            if local_val != remote_val:
                ret["changes"] = {"old": remote_configs, "new": local_configs}
                return ret

    return ret



import os

import salt.modules.file as file_
import salt.modules.heat as heat
import salt.modules.win_file as win_file
import salt.utils.platform
import salt.utils.win_dacl as dacl
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class MockStacks:
    """
    Mock stacks.StackManager
    """

    def validate(self, **kwargs):
        """
        Mock of stacks.StackManager.validate method
        """
        self.mock_val_ret = MagicMock()
        self.mock_val_ret.json.return_value = {"result": "mocked response"}
        self.mock_validate = MagicMock()
        self.mock_validate.post.return_value = self.mock_val_ret
        return self.mock_validate

    def create(self, **kwargs):
        self.mock_create_ret = MagicMock()
        self.mock_create_ret.json.return_value = {
            "result": "mocked create",
            "fields": kwargs,
        }
        self.mock_create = MagicMock()
        self.mock_create.post.return_value = self.mock_create_ret
        return self.mock_create

    def update(self, name, **kwargs):
        self.mock_update_ret = MagicMock()
        self.mock_update_ret.json.return_value = {
            "result": "mocked update",
            "fields": kwargs,
            "name": name,
        }
        self.mock_update = MagicMock()
        self.mock_update.post.return_value = self.mock_update_ret
        return self.mock_update


class MockClient:
    """
    Mock of Client class
    """

    def __init__(self, profile=None, **conn_args):
        self.stacks = MockStacks()


class HeatTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.heat
    """

    def setup_loader_modules(self):
        return {
            heat: {"_auth": MockClient},
            file_: {
                "__opts__": {
                    "hash_type": "sha256",
                    "cachedir": os.path.join(RUNTIME_VARS.TMP, "rootdir", "cache"),
                    "test": False,
                },
                "__salt__": {
                    "config.option": MagicMock(
                        return_value={"obfuscate_templates": False}
                    ),
                    "config.backup_mode": MagicMock(return_value=False),
                },
            },
            win_file: {
                "__utils__": {"dacl.check_perms": salt.utils.win_dacl.check_perms}
            },
            dacl: {"__opts__": {"test": False}},
        }

    def setUp(self):
        self.patch_check = patch("salt.modules.file.check_perms", file_.check_perms)
        if salt.utils.platform.is_windows():
            self.patch_check = patch(
                "salt.modules.file.check_perms", win_file.check_perms
            )

    def test_heat_create_stack(self):
        """
        Test salt.modules.heat.create_stack method
        """
        patch_file = patch.dict(
            heat.__salt__,
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )

        with patch_file, self.patch_check:
            ret = heat.create_stack(
                name="mystack",
                profile="openstack1",
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
            )
        assert ret == {"result": True, "comment": "Created stack 'mystack'."}

    def test_heat_create_stack_environment(self):
        """
        Test salt.modules.heat.create_stack method with environment set
        """
        patch_file = patch.dict(
            "salt.modules.heat.__salt__",
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )
        with patch_file, self.patch_check:
            ret = heat.create_stack(
                name="mystack",
                profile="openstack1",
                environment=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml"
                ),
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
            )
        assert ret == {"result": True, "comment": "Created stack 'mystack'."}

    def test_heat_create_stack_environment_err(self):
        """
        Test salt.modules.heat.create_stack method with environment set
        and there is an error reading the environment file
        """
        patch_file = patch.dict(
            "salt.modules.heat.__salt__",
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": MagicMock(
                    side_effect=[{"result": True}, {"result": False}]
                ),
            },
        )
        patch_template = patch(
            "salt.modules.heat._parse_template", MagicMock(return_value=True)
        )
        env_file = os.path.join(RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml")
        with patch_file, patch_template, self.patch_check:
            ret = heat.create_stack(
                name="mystack",
                profile="openstack1",
                environment=env_file,
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
            )
        assert ret == {
            "result": False,
            "comment": "Can not open environment: {}, ".format(env_file),
        }

    def test_heat_update_stack(self):
        """
        Test salt.modules.heat.update_method method
        """
        patch_file = patch.dict(
            heat.__salt__,
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )
        with patch_file, self.patch_check:
            ret = heat.update_stack(
                name="mystack",
                profile="openstack1",
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
            )
        assert ret == {"result": True, "comment": ("Updated stack 'mystack'.",)}

    def test_heat_update_stack_env(self):
        """
        Test salt.modules.heat.update_method method
        with environment set
        """
        patch_file = patch.dict(
            heat.__salt__,
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )
        with patch_file, self.patch_check:
            ret = heat.update_stack(
                name="mystack",
                profile="openstack1",
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
                environment=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml"
                ),
            )
        assert ret == {"result": True, "comment": ("Updated stack 'mystack'.",)}

    def test_heat_update_stack_env_err(self):
        """
        Test salt.modules.heat.update_method method
        with environment set and there is an error
        reading the environment file
        """
        patch_file = patch.dict(
            heat.__salt__,
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": MagicMock(
                    side_effect=[{"result": True}, {"result": False}]
                ),
            },
        )
        with patch_file, self.patch_check:
            ret = heat.update_stack(
                name="mystack",
                profile="openstack1",
                template_file=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
                environment=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml"
                ),
            )
        assert ret == {
            "result": False,
            "comment": "Error parsing template Template format version not found.",
        }



import time
from multiprocessing import Manager, Process

import psutil
import pytest

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture
def testfile_path(tmp_path):
    return tmp_path / "testfile"


@pytest.fixture
def file_add_delete_sls(testfile_path, base_env_state_tree_root_dir):
    sls_name = "file_add"
    sls_contents = """
    add_file:
      file.managed:
        - name: {path}
        - source: salt://testfile
        - makedirs: true
        - require:
          - cmd: echo

    delete_file:
      file.absent:
        - name: {path}
        - require:
          - file: add_file

    echo:
      cmd.run:
        - name: \"echo 'This is a test!'\"
    """.format(
        path=testfile_path
    )
    with pytest.helpers.temp_file(
        "{}.sls".format(sls_name), sls_contents, base_env_state_tree_root_dir
    ):
        yield sls_name


@pytest.mark.skip_on_darwin(reason="MacOS is a spawning platform, won't work")
@pytest.mark.flaky(max_runs=4)
def test_memory_leak(salt_cli, salt_minion, file_add_delete_sls):
    max_usg = None

    # Using shared variables to be able to send a stop flag to the process
    with Manager() as manager:
        done_flag = manager.list()
        during_run_data = manager.list()

        def _func(data, flag):
            while len(flag) == 0:
                time.sleep(0.05)
                usg = psutil.virtual_memory()
                data.append(usg.total - usg.available)

        proc = Process(target=_func, args=(during_run_data, done_flag))
        proc.start()

        # Try to drive up memory usage
        for _ in range(50):
            salt_cli.run("state.sls", file_add_delete_sls, minion_tgt=salt_minion.id)

        done_flag.append(1)
        proc.join()

        start_usg = during_run_data[0]
        max_usg = during_run_data[0]
        for row in during_run_data[1:]:
            max_usg = row if row >= max_usg else max_usg

    # This would be weird, but should account for it
    if max_usg > start_usg:
        max_tries = 50
        # The maximum that the current usage can be in order to pass the test
        threshold = (max_usg - start_usg) * 0.25 + start_usg
        for _ in range(max_tries):
            usg = psutil.virtual_memory()
            current_usg = usg.total - usg.available
            if current_usg <= start_usg:
                break
            # Get percent difference between max and start usg that current usg is at
            if current_usg <= threshold:
                break

            time.sleep(2)
        else:
            pytest.fail("Memory usage did not drop off appropriately")



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.http
"""


import pytest

import salt.modules.http as http
import salt.utils.http
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {http: {}}


def test_query():
    """
    Test for Query a resource, and decode the return data
    """
    with patch.object(salt.utils.http, "query", return_value="A"):
        assert http.query("url") == "A"


def test_wait_for_with_interval():
    """
    Test for wait_for_successful_query waits for request_interval
    """

    query_mock = MagicMock(side_effect=[{"error": "error"}, {}])

    with patch.object(salt.utils.http, "query", query_mock):
        with patch("time.sleep", MagicMock()) as sleep_mock:
            assert http.wait_for_successful_query("url", request_interval=1) == {}
            sleep_mock.assert_called_once_with(1)


def test_wait_for_without_interval():
    """
    Test for wait_for_successful_query waits for request_interval
    """

    query_mock = MagicMock(side_effect=[{"error": "error"}, {}])

    with patch.object(salt.utils.http, "query", query_mock):
        with patch("time.sleep", MagicMock()) as sleep_mock:
            assert http.wait_for_successful_query("url") == {}
            sleep_mock.assert_not_called()



"""
Module to provide Postgres compatibility to salt for debian family specific tools.

"""
import logging
import shlex

import salt.utils.path

log = logging.getLogger(__name__)

__virtualname__ = "postgres"


def __virtual__():
    """
    Only load this module if the pg_createcluster bin exists
    """
    if salt.utils.path.which("pg_createcluster"):
        return __virtualname__
    return (
        False,
        "postgres execution module not loaded: pg_createcluste command not found.",
    )


def cluster_create(
    version,
    name="main",
    port=None,
    locale=None,
    encoding=None,
    datadir=None,
    allow_group_access=None,
    data_checksums=None,
    wal_segsize=None,
):
    """
    Adds a cluster to the Postgres server.

    .. warning:

       Only works for debian family distros so far.

    CLI Example:

    .. code-block:: bash

        salt '*' postgres.cluster_create '9.3'

        salt '*' postgres.cluster_create '9.3' 'main'

        salt '*' postgres.cluster_create '9.3' locale='fr_FR'

        salt '*' postgres.cluster_create '11' data_checksums=True wal_segsize='32'
    """

    cmd = [salt.utils.path.which("pg_createcluster")]
    if port:
        cmd += ["--port", str(port)]
    if locale:
        cmd += ["--locale", locale]
    if encoding:
        cmd += ["--encoding", encoding]
    if datadir:
        cmd += ["--datadir", datadir]
    cmd += [str(version), name]
    # initdb-specific options are passed after '--'
    if allow_group_access or data_checksums or wal_segsize:
        cmd += ["--"]
    if allow_group_access is True:
        cmd += ["--allow-group-access"]
    if data_checksums is True:
        cmd += ["--data-checksums"]
    if wal_segsize:
        cmd += ["--wal-segsize", wal_segsize]
    cmdstr = " ".join([shlex.quote(c) for c in cmd])
    ret = __salt__["cmd.run_all"](cmdstr, python_shell=False)
    if ret.get("retcode", 0) != 0:
        log.error("Error creating a Postgresql cluster %s/%s", version, name)
        return False
    return ret


def cluster_list(verbose=False):
    """
    Return a list of cluster of Postgres server (tuples of version and name).

    CLI Example:

    .. code-block:: bash

        salt '*' postgres.cluster_list

        salt '*' postgres.cluster_list verbose=True
    """
    cmd = [salt.utils.path.which("pg_lsclusters"), "--no-header"]
    ret = __salt__["cmd.run_all"](" ".join([shlex.quote(c) for c in cmd]))
    if ret.get("retcode", 0) != 0:
        log.error("Error listing clusters")
    cluster_dict = _parse_pg_lscluster(ret["stdout"])
    if verbose:
        return cluster_dict
    return cluster_dict.keys()


def cluster_exists(version, name="main"):
    """
    Checks if a given version and name of a cluster exists.

    CLI Example:

    .. code-block:: bash

        salt '*' postgres.cluster_exists '9.3'

        salt '*' postgres.cluster_exists '9.3' 'main'
    """
    return f"{version}/{name}" in cluster_list()


def cluster_remove(version, name="main", stop=False):
    """
    Remove a cluster on a Postgres server. By default it doesn't try
    to stop the cluster.

    CLI Example:

    .. code-block:: bash

        salt '*' postgres.cluster_remove '9.3'

        salt '*' postgres.cluster_remove '9.3' 'main'

        salt '*' postgres.cluster_remove '9.3' 'main' stop=True

    """
    cmd = [salt.utils.path.which("pg_dropcluster")]
    if stop:
        cmd += ["--stop"]
    cmd += [str(version), name]
    cmdstr = " ".join([shlex.quote(c) for c in cmd])
    ret = __salt__["cmd.run_all"](cmdstr, python_shell=False)
    # FIXME - return Boolean ?
    if ret.get("retcode", 0) != 0:
        log.error("Error removing a Postgresql cluster %s/%s", version, name)
    else:
        ret["changes"] = f"Successfully removed cluster {version}/{name}"
    return ret


def _parse_pg_lscluster(output):
    """
    Helper function to parse the output of pg_lscluster
    """
    cluster_dict = {}
    for line in output.splitlines():
        version, name, port, status, user, datadir, log = line.split()
        cluster_dict[f"{version}/{name}"] = {
            "port": int(port),
            "status": status,
            "user": user,
            "datadir": datadir,
            "log": log,
        }
    return cluster_dict



"""
Module for Management of Memcached Keys

.. versionadded:: 2014.1.0
"""

import logging

import salt.utils.functools
from salt.exceptions import CommandExecutionError, SaltInvocationError

# TODO: use salt.utils.memcache


try:
    import memcache

    HAS_MEMCACHE = True
except ImportError:
    HAS_MEMCACHE = False

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 11211
DEFAULT_TIME = 0
DEFAULT_MIN_COMPRESS_LEN = 0

log = logging.getLogger(__name__)

# Don't shadow built-ins
__func_alias__ = {"set_": "set"}

__virtualname__ = "memcached"


def __virtual__():
    """
    Only load if python-memcache is installed
    """
    if HAS_MEMCACHE:
        return __virtualname__
    return (
        False,
        "The memcached execution module cannot be loaded: "
        "python memcache library not available.",
    )


def _connect(host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Returns a tuple of (user, host, port) with config, pillar, or default
    values assigned to missing values.
    """
    if str(port).isdigit():
        return memcache.Client(["{}:{}".format(host, port)], debug=0)
    raise SaltInvocationError("port must be an integer")


def _check_stats(conn):
    """
    Helper function to check the stats data passed into it, and raise an
    exception if none are returned. Otherwise, the stats are returned.
    """
    stats = conn.get_stats()
    if not stats:
        raise CommandExecutionError("memcached server is down or does not exist")
    return stats


def status(host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Get memcached status

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.status
    """
    conn = _connect(host, port)
    try:
        stats = _check_stats(conn)[0]
    except (CommandExecutionError, IndexError):
        return False
    else:
        return {stats[0]: stats[1]}


def get(key, host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Retrieve value for a key

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.get <key>
    """
    conn = _connect(host, port)
    _check_stats(conn)
    return conn.get(key)


def set_(
    key,
    value,
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    time=DEFAULT_TIME,
    min_compress_len=DEFAULT_MIN_COMPRESS_LEN,
):
    """
    Set a key on the memcached server, overwriting the value if it exists.

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.set <key> <value>
    """
    if not isinstance(time, int):
        raise SaltInvocationError("'time' must be an integer")
    if not isinstance(min_compress_len, int):
        raise SaltInvocationError("'min_compress_len' must be an integer")
    conn = _connect(host, port)
    _check_stats(conn)
    return conn.set(key, value, time, min_compress_len)


def delete(key, host=DEFAULT_HOST, port=DEFAULT_PORT, time=DEFAULT_TIME):
    """
    Delete a key from memcache server

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.delete <key>
    """
    if not isinstance(time, int):
        raise SaltInvocationError("'time' must be an integer")
    conn = _connect(host, port)
    _check_stats(conn)
    return bool(conn.delete(key, time))


def add(
    key,
    value,
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    time=DEFAULT_TIME,
    min_compress_len=DEFAULT_MIN_COMPRESS_LEN,
):
    """
    Add a key to the memcached server, but only if it does not exist. Returns
    False if the key already exists.

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.add <key> <value>
    """
    if not isinstance(time, int):
        raise SaltInvocationError("'time' must be an integer")
    if not isinstance(min_compress_len, int):
        raise SaltInvocationError("'min_compress_len' must be an integer")
    conn = _connect(host, port)
    _check_stats(conn)
    return conn.add(key, value, time=time, min_compress_len=min_compress_len)


def replace(
    key,
    value,
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    time=DEFAULT_TIME,
    min_compress_len=DEFAULT_MIN_COMPRESS_LEN,
):
    """
    Replace a key on the memcached server. This only succeeds if the key
    already exists. This is the opposite of :mod:`memcached.add
    <salt.modules.memcached.add>`

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.replace <key> <value>
    """
    if not isinstance(time, int):
        raise SaltInvocationError("'time' must be an integer")
    if not isinstance(min_compress_len, int):
        raise SaltInvocationError("'min_compress_len' must be an integer")
    conn = _connect(host, port)
    stats = conn.get_stats()
    return conn.replace(key, value, time=time, min_compress_len=min_compress_len)


def increment(key, delta=1, host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Increment the value of a key

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.increment <key>
        salt '*' memcached.increment <key> 2
    """
    conn = _connect(host, port)
    _check_stats(conn)
    cur = get(key)

    if cur is None:
        raise CommandExecutionError("Key '{}' does not exist".format(key))
    elif not isinstance(cur, int):
        raise CommandExecutionError(
            "Value for key '{}' must be an integer to be incremented".format(key)
        )

    try:
        return conn.incr(key, delta)
    except ValueError:
        raise SaltInvocationError("Delta value must be an integer")


incr = salt.utils.functools.alias_function(increment, "incr")


def decrement(key, delta=1, host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Decrement the value of a key

    CLI Example:

    .. code-block:: bash

        salt '*' memcached.decrement <key>
        salt '*' memcached.decrement <key> 2
    """
    conn = _connect(host, port)
    _check_stats(conn)

    cur = get(key)
    if cur is None:
        raise CommandExecutionError("Key '{}' does not exist".format(key))
    elif not isinstance(cur, int):
        raise CommandExecutionError(
            "Value for key '{}' must be an integer to be decremented".format(key)
        )

    try:
        return conn.decr(key, delta)
    except ValueError:
        raise SaltInvocationError("Delta value must be an integer")


decr = salt.utils.functools.alias_function(decrement, "decr")



from salt.utils.sanitizers import clean, mask_args_value
from tests.support.unit import TestCase


class SanitizersTestCase(TestCase):
    """
    TestCase for sanitizers
    """

    def test_sanitized_trim(self):
        """
        Test sanitized input for trimming
        """
        value = " sample "
        response = clean.trim(value)
        assert response == "sample"
        assert type(response) == str

    def test_sanitized_filename(self):
        """
        Test sanitized input for filename
        """
        value = "/absolute/path/to/the/file.txt"
        response = clean.filename(value)
        assert response == "file.txt"

        value = "../relative/path/to/the/file.txt"
        response = clean.filename(value)
        assert response == "file.txt"

    def test_sanitized_hostname(self):
        """
        Test sanitized input for hostname (id)
        """
        value = "   ../ ../some/dubious/hostname      "
        response = clean.hostname(value)
        assert response == "somedubioushostname"

    test_sanitized_id = test_sanitized_hostname

    def test_value_masked(self):
        """
        Test if the values are masked.
        :return:
        """
        out = mask_args_value("quantum: fluctuations", "quant*")
        assert out == "quantum: ** hidden **"

    def test_value_not_masked(self):
        """
        Test if the values are not masked.
        :return:
        """
        out = mask_args_value("quantum fluctuations", "quant*")
        assert out == "quantum fluctuations"



import pytest

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_requisites_full_sls_prereq(state, state_tree):
    """
    Test the sls special command in requisites
    """
    full_sls_contents = """
    B:
      cmd.run:
        - name: echo B
    C:
      cmd.run:
        - name: echo C
    """
    sls_contents = """
    include:
      - fullsls
    A:
      cmd.run:
        - name: echo A
        - prereq:
          - sls: fullsls
    """
    expected_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo C" run',
            "result": True,
            "changes": True,
        },
    }
    with pytest.helpers.temp_file(
        "requisite.sls", sls_contents, state_tree
    ), pytest.helpers.temp_file("fullsls.sls", full_sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_1(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # B --+
    #     |
    # C <-+ ----+
    #           |
    # A <-------+

    # runs after C
    A:
      cmd.run:
        - name: echo A third
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C

    # runs before C
    B:
      cmd.run:
        - name: echo B first
        # will test C and be applied only if C changes,
        # and then will run before C
        - prereq:
          - cmd: C
    C:
      cmd.run:
        - name: echo C second

    # will fail with "The following requisites were not found"
    I:
      cmd.run:
        - name: echo I
        - prereq:
          - cmd: Z
    J:
      cmd.run:
        - name: echo J
        - prereq:
          - foobar: A
    """
    expected_result = {
        "cmd_|-A_|-echo A third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo A third" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo C second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-I_|-echo I_|-run": {
            "__run_num__": 3,
            "comment": "The following requisites were not found:\n"
            + "                   prereq:\n"
            + "                       cmd: Z\n",
            "result": False,
            "changes": False,
        },
        "cmd_|-J_|-echo J_|-run": {
            "__run_num__": 4,
            "comment": "The following requisites were not found:\n"
            + "                   prereq:\n"
            + "                       foobar: A\n",
            "result": False,
            "changes": False,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_2(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    # same test, but not using lists in yaml syntax
    sls_contents = """
    # B --+
    #     |
    # C <-+ ----+
    #           |
    # A <-------+

    # runs after C
    A:
      cmd.run:
        - name: echo A third
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
           cmd: C

    # runs before C
    B:
      cmd.run:
        - name: echo B first
        # will test C and be applied only if C changes,
        # and then will run before C
        - prereq:
            cmd: C
    C:
      cmd.run:
        - name: echo C second
    """
    errmsg = (
        "The prereq statement in state 'B' in SLS 'requisite' needs to be formed as a"
        " list"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_prereq_simple_ordering_and_errors_3(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # B --+
    #     |
    # C <-+ ----+
    #           |
    # A <-------+

    # runs after C
    A:
      cmd.run:
        - name: echo A third
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - C

    # runs before C
    B:
      cmd.run:
        - name: echo B first
        # will test C and be applied only if C changes,
        # and then will run before C
        - prereq:
          - C
    C:
      cmd.run:
        - name: echo C second

    # will fail with "The following requisites were not found"
    I:
      cmd.run:
        - name: echo I
        - prereq:
          - Z
        """
    expected_result = {
        "cmd_|-A_|-echo A third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo A third" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo C second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-I_|-echo I_|-run": {
            "__run_num__": 3,
            "comment": "The following requisites were not found:\n"
            + "                   prereq:\n"
            + "                       id: Z\n",
            "result": False,
            "changes": False,
        },
    }

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_4(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # Theory:
    #
    # C <--+ <--+ <-+ <-+
    #      |    |   |   |
    # A ---+    |   |   |
    #           |   |   |
    # B --------+   |   |
    #               |   |
    # D-------------+   |
    #                   |
    # E-----------------+

    # runs after C
    A:
      cmd.run:
        - name: echo A
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C

    B:
      cmd.run:
        - name: echo B

    # runs before D and B
    C:
      cmd.run:
        - name: echo C
        # will test D and be applied only if D changes,
        # and then will run before D. Same for B
        - prereq:
          - cmd: B
          - cmd: D

    D:
      cmd.run:
        - name: echo D

    E:
      cmd.run:
        - name: echo E
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C
    """
    expected_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo C" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-D_|-echo D_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo D" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-E_|-echo E_|-run": {
            "__run_num__": 4,
            "comment": 'Command "echo E" run',
            "result": True,
            "changes": True,
        },
    }

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_5(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # A --+
    #     |
    # B <-+ ----+
    #           |
    # C <-------+

    # runs before A and/or B
    A:
      cmd.run:
        - name: echo A first
        # is running in test mode before B/C
        - prereq:
          - cmd: B
          - cmd: C

    # always has to run
    B:
      cmd.run:
        - name: echo B second

    # never has to run
    C:
      cmd.wait:
        - name: echo C third
    """
    expected_result = {
        "cmd_|-A_|-echo A first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo A first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo B second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C third_|-wait": {
            "__run_num__": 2,
            "comment": "",
            "result": True,
            "changes": False,
        },
    }

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_6(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # issue #8211
    #             expected rank
    # B --+             1
    #     |
    # C <-+ ----+       2/3
    #           |
    # D ---+    |       3/2
    #      |    |
    # A <--+ <--+       4
    #
    #             resulting rank
    # D --+
    #     |
    # A <-+ <==+
    #          |
    # B --+    +--> unrespected A prereq_in C (FAILURE)
    #     |    |
    # C <-+ ===+

    # runs after C
    A:
      cmd.run:
        - name: echo A fourth
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C

    # runs before C
    B:
      cmd.run:
        - name: echo B first
        # will test C and be applied only if C changes,
        # and then will run before C
        - prereq:
          - cmd: C

    C:
      cmd.run:
        - name: echo C second
        # replacing A prereq_in C by theses lines
        # changes nothing actually
        #- prereq:
        #  - cmd: A

    # Removing D, A gets executed after C
    # as described in (A prereq_in C)
    # runs before A
    D:
      cmd.run:
        - name: echo D third
        # will test A and be applied only if A changes,
        # and then will run before A
        - prereq:
          - cmd: A
    """
    expected_result = {
        "cmd_|-A_|-echo A fourth_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo A fourth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo C second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-D_|-echo D third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo D third" run',
            "result": True,
            "changes": True,
        },
    }

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_prereq_simple_ordering_and_errors_7(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # will fail with 'Cannot extend ID Z (...) not part of the high state.'
    # and not "The following requisites were not found" like in yaml list syntax
    I:
      cmd.run:
        - name: echo I
        - prereq:
          - cmd: Z
    """
    errmsg = (
        "The following requisites were not found:\n"
        "                   prereq:\n"
        "                       cmd: Z\n"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["cmd_|-I_|-echo I_|-run"].comment == errmsg


def test_requisites_prereq_simple_ordering_and_errors_8(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.run:
        - name: echo A

    B:
      cmd.run:
        - name: echo B
        - prereq:
          - foobar: A
    """
    errmsg = (
        "The following requisites were not found:\n"
        "                   prereq:\n"
        "                       foobar: A\n"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["cmd_|-B_|-echo B_|-run"].comment == errmsg


def test_requisites_prereq_simple_ordering_and_errors_9(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.run:
        - name: echo A

    B:
      cmd.run:
        - name: echo B
        - prereq:
          - foobar: C
    """
    errmsg = (
        "The following requisites were not found:\n"
        "                   prereq:\n"
        "                       foobar: C\n"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["cmd_|-B_|-echo B_|-run"].comment == errmsg


@pytest.mark.skip("issue #8210 : prereq recursion undetected")
def test_requisites_prereq_simple_ordering_and_errors_10(state, state_tree):
    """
    Call sls file containing several prereq_in and prereq.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.run:
        - name: echo A
        - prereq_in:
          - cmd: B
    B:
      cmd.run:
        - name: echo B
        - prereq_in:
          - cmd: A
    """
    errmsg = (
        'A recursive requisite was found, SLS "requisites.prereq_recursion_error" ID'
        ' "B" ID "A"'
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_infinite_recursion_sls_prereq(state, state_tree):
    sls_contents = """
    include:
      - requisite2
    A:
      test.succeed_without_changes:
        - name: A
        - prereq:
          - sls: requisite2
    """
    sls_2_contents = """
    B:
      test.succeed_without_changes:
        - name: B
    """
    with pytest.helpers.temp_file(
        "requisite.sls", sls_contents, state_tree
    ), pytest.helpers.temp_file("requisite2.sls", sls_2_contents, state_tree):
        ret = state.sls("requisite")
        for state_return in ret:
            assert state_return.result is True


def test_infinite_recursion_prereq(state, state_tree):
    sls_contents = """
    A:
      test.nop:
        - prereq:
          - test: B
    B:
      test.nop:
        - require:
          - name: non-existant
    C:
      test.nop:
        - require:
          - test: B
    """

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        for state_return in ret:
            assert state_return.result is False


def test_infinite_recursion_prereq2(state, state_tree):
    sls_contents = """
    A:
      test.nop:
        - prereq:
          - test: B
    B:
      test.nop:
        - require:
          - test: D
    C:
      test.nop:
        - require:
          - test: B
    D:
      test.nop: []
    """

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        for state_return in ret:
            assert state_return.result is True


def test_requisites_prereq_fail_in_prereq(state, state_tree):
    sls_contents = """
    State A:
      test.configurable_test_state:
        - result: True
        - changes: True
        - name: fail

    State B:
      test.configurable_test_state:
        - changes: True
        - result: False
        - prereq:
          - test: State A

    State C:
      test.nop:
        - onchanges:
          - test: State A
    """

    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["test_|-State A_|-fail_|-configurable_test_state"].result is None
        assert (
            ret["test_|-State A_|-fail_|-configurable_test_state"].full_return[
                "changes"
            ]
            == {}
        )

        assert not ret["test_|-State B_|-State B_|-configurable_test_state"].result

        assert ret["test_|-State C_|-State C_|-nop"].result
        assert not ret["test_|-State C_|-State C_|-nop"].full_return["__state_ran__"]
        assert (
            ret["test_|-State C_|-State C_|-nop"].full_return["comment"]
            == "State was not run because none of the onchanges reqs changed"
        )



"""
Unit Tests for the k8s execution module.
"""

import base64
import hashlib
import time
from subprocess import PIPE, Popen

import pytest

import salt.modules.k8s as k8s
import salt.utils.files
import salt.utils.json
from tests.support.unit import TestCase


@pytest.mark.skip_if_binaries_missing("kubectl")
class TestK8SNamespace(TestCase):

    maxDiff = None

    def test_get_namespaces(self):
        res = k8s.get_namespaces(apiserver_url="http://127.0.0.1:8080")
        a = len(res.get("items"))
        proc = Popen(["kubectl", "get", "namespaces", "-o", "json"], stdout=PIPE)
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = len(kubectl_out.get("items"))
        self.assertEqual(a, b)

    def test_get_one_namespace(self):
        res = k8s.get_namespaces("default", apiserver_url="http://127.0.0.1:8080")
        a = res.get("metadata", {}).get("name", "a")
        proc = Popen(
            ["kubectl", "get", "namespaces", "default", "-o", "json"], stdout=PIPE
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = kubectl_out.get("metadata", {}).get("name", "b")
        self.assertEqual(a, b)

    def test_create_namespace(self):
        hash = hashlib.sha1()
        hash.update(str(time.time()))
        nsname = hash.hexdigest()[:16]
        res = k8s.create_namespace(nsname, apiserver_url="http://127.0.0.1:8080")
        proc = Popen(
            ["kubectl", "get", "namespaces", nsname, "-o", "json"], stdout=PIPE
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        # if creation is failed, kubernetes return non json error message
        self.assertTrue(isinstance(kubectl_out, dict))


@pytest.mark.skip_if_binaries_missing("kubectl")
class TestK8SSecrets(TestCase):

    maxDiff = None

    def setUp(self):
        hash = hashlib.sha1()
        hash.update(str(time.time()))
        self.name = hash.hexdigest()[:16]
        data = {"testsecret": base64.encodestring("teststring")}
        self.request = {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {"name": self.name, "namespace": "default"},
            "data": data,
        }

    def test_get_secrets(self):
        res = k8s.get_secrets("default", apiserver_url="http://127.0.0.1:8080")
        a = len(res.get("items", []))
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "secrets", "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = len(kubectl_out.get("items", []))
        self.assertEqual(a, b)

    def test_get_one_secret(self):
        name = self.name
        filename = "/tmp/{}.json".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            salt.utils.json.dump(self.request, f)

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.1)
        res = k8s.get_secrets("default", name, apiserver_url="http://127.0.0.1:8080")
        a = res.get("metadata", {}).get("name", "a")
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "secrets", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = kubectl_out.get("metadata", {}).get("name", "b")
        self.assertEqual(a, b)

    def test_get_decoded_secret(self):
        name = self.name
        filename = "/tmp/{}.json".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            salt.utils.json.dump(self.request, f)

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give etcd to populate data on all nodes
        time.sleep(0.1)
        res = k8s.get_secrets(
            "default", name, apiserver_url="http://127.0.0.1:8080", decode=True
        )
        a = res.get("data", {}).get(
            "testsecret",
        )
        self.assertEqual(a, "teststring")

    def test_create_secret(self):
        name = self.name
        names = []
        expected_data = {}
        for i in range(2):
            names.append("/tmp/{}-{}".format(name, i))
            with salt.utils.files.fopen("/tmp/{}-{}".format(name, i), "w") as f:
                expected_data["{}-{}".format(name, i)] = base64.b64encode(
                    "{}{}".format(name, i)
                )
                f.write(salt.utils.stringutils.to_str("{}{}".format(name, i)))
        res = k8s.create_secret(
            "default", name, names, apiserver_url="http://127.0.0.1:8080"
        )
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "secrets", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        # if creation is failed, kubernetes return non json error message
        b = kubectl_out.get("data", {})
        self.assertTrue(isinstance(kubectl_out, dict))
        self.assertEqual(expected_data, b)

    def test_update_secret(self):
        name = self.name
        filename = "/tmp/{}.json".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            salt.utils.json.dump(self.request, f)

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.1)
        expected_data = {}
        names = []
        for i in range(3):
            names.append("/tmp/{}-{}-updated".format(name, i))
            with salt.utils.files.fopen("/tmp/{}-{}-updated".format(name, i), "w") as f:
                expected_data["{}-{}-updated".format(name, i)] = base64.b64encode(
                    "{}{}-updated".format(name, i)
                )
                f.write("{}{}-updated".format(name, i))

        res = k8s.update_secret(
            "default", name, names, apiserver_url="http://127.0.0.1:8080"
        )
        # if creation is failed, kubernetes return non json error message
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "secrets", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        # if creation is failed, kubernetes return non json error message
        b = kubectl_out.get("data", {})
        self.assertTrue(isinstance(kubectl_out, dict))
        self.assertEqual(expected_data, b)

    def test_delete_secret(self):
        name = self.name
        filename = "/tmp/{}.json".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            salt.utils.json.dump(self.request, f)

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.1)
        res = k8s.delete_secret("default", name, apiserver_url="http://127.0.0.1:8080")
        time.sleep(0.1)
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "secrets", name, "-o", "json"],
            stdout=PIPE,
            stderr=PIPE,
        )
        kubectl_out, err = proc.communicate()
        # stdout is empty, stderr is showing something like "not found"
        self.assertEqual("", kubectl_out)
        self.assertEqual(
            'Error from server: secrets "{}" not found\n'.format(name), err
        )


@pytest.mark.skip_if_binaries_missing("kubectl")
class TestK8SResourceQuotas(TestCase):

    maxDiff = None

    def setUp(self):
        hash = hashlib.sha1()
        hash.update(str(time.time()))
        self.name = hash.hexdigest()[:16]

    def test_get_resource_quotas(self):
        name = self.name
        namespace = self.name
        create_namespace = Popen(
            ["kubectl", "create", "namespace", namespace], stdout=PIPE
        )
        create_namespace = Popen(
            ["kubectl", "create", "namespace", namespace], stdout=PIPE
        )
        request = """
apiVersion: v1
kind: ResourceQuota
metadata:
  name: {}
spec:
  hard:
    cpu: "20"
    memory: 1Gi
    persistentvolumeclaims: "10"
    pods: "10"
    replicationcontrollers: "20"
    resourcequotas: "1"
    secrets: "10"
    services: "5"
""".format(
            name
        )
        filename = "/tmp/{}.yaml".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            f.write(salt.utils.stringutils.to_str(request))

        create = Popen(
            ["kubectl", "--namespace={}".format(namespace), "create", "-f", filename],
            stdout=PIPE,
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.2)
        res = k8s.get_resource_quotas(namespace, apiserver_url="http://127.0.0.1:8080")
        a = len(res.get("items", []))
        proc = Popen(
            [
                "kubectl",
                "--namespace={}".format(namespace),
                "get",
                "quota",
                "-o",
                "json",
            ],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = len(kubectl_out.get("items", []))
        self.assertEqual(a, b)

    def test_get_one_resource_quota(self):
        name = self.name
        namespace = self.name
        create_namespace = Popen(
            ["kubectl", "create", "namespace", namespace], stdout=PIPE
        )
        request = """
apiVersion: v1
kind: ResourceQuota
metadata:
  name: {}
spec:
  hard:
    cpu: "20"
    memory: 1Gi
    persistentvolumeclaims: "10"
    pods: "10"
    replicationcontrollers: "20"
    resourcequotas: "1"
    secrets: "10"
    services: "5"
""".format(
            name
        )
        filename = "/tmp/{}.yaml".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            f.write(salt.utils.stringutils.to_str(request))

        create = Popen(
            ["kubectl", "--namespace={}".format(namespace), "create", "-f", filename],
            stdout=PIPE,
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.2)
        res = k8s.get_resource_quotas(
            namespace, name, apiserver_url="http://127.0.0.1:8080"
        )
        a = res.get("metadata", {}).get("name", "a")
        proc = Popen(
            [
                "kubectl",
                "--namespace={}".format(namespace),
                "get",
                "quota",
                name,
                "-o",
                "json",
            ],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = kubectl_out.get("metadata", {}).get("name", "b")
        self.assertEqual(a, b)

    def test_create_resource_quota(self):
        name = self.name
        namespace = self.name
        create_namespace = Popen(
            ["kubectl", "create", "namespace", namespace], stdout=PIPE
        )
        quota = {"cpu": "20", "memory": "1Gi"}
        res = k8s.create_resource_quota(
            namespace, quota, name=name, apiserver_url="http://127.0.0.1:8080"
        )
        proc = Popen(
            [
                "kubectl",
                "--namespace={}".format(namespace),
                "get",
                "quota",
                name,
                "-o",
                "json",
            ],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        self.assertTrue(isinstance(kubectl_out, dict))

    def test_update_resource_quota(self):
        name = self.name
        namespace = self.name
        create_namespace = Popen(
            ["kubectl", "create", "namespace", namespace], stdout=PIPE
        )
        request = """
apiVersion: v1
kind: ResourceQuota
metadata:
  name: {}
spec:
  hard:
    cpu: "20"
    memory: 1Gi
    persistentvolumeclaims: "10"
    pods: "10"
    replicationcontrollers: "20"
    resourcequotas: "1"
    secrets: "10"
    services: "5"
""".format(
            name
        )
        filename = "/tmp/{}.yaml".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            f.write(salt.utils.stringutils.to_str(request))

        create = Popen(
            ["kubectl", "--namespace={}".format(namespace), "create", "-f", filename],
            stdout=PIPE,
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.2)
        quota = {"cpu": "10", "memory": "2Gi"}
        res = k8s.create_resource_quota(
            namespace,
            quota,
            name=name,
            apiserver_url="http://127.0.0.1:8080",
            update=True,
        )
        proc = Popen(
            [
                "kubectl",
                "--namespace={}".format(namespace),
                "get",
                "quota",
                name,
                "-o",
                "json",
            ],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        limit = kubectl_out.get("spec").get("hard").get("memory")
        self.assertEqual("2Gi", limit)


@pytest.mark.skip_if_binaries_missing("kubectl")
class TestK8SLimitRange(TestCase):

    maxDiff = None

    def setUp(self):
        hash = hashlib.sha1()
        hash.update(str(time.time()))
        self.name = hash.hexdigest()[:16]

    def test_create_limit_range(self):
        name = self.name
        limits = {"Container": {"defaultRequest": {"cpu": "100m"}}}
        res = k8s.create_limit_range(
            "default", limits, name=name, apiserver_url="http://127.0.0.1:8080"
        )
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "limits", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        self.assertTrue(isinstance(kubectl_out, dict))

    def test_update_limit_range(self):
        name = self.name
        request = """
apiVersion: v1
kind: LimitRange
metadata:
  name: {}
spec:
  limits:
  - default:
      cpu: 200m
      memory: 512Mi
    defaultRequest:
      cpu: 100m
      memory: 256Mi
    type: Container
""".format(
            name
        )
        limits = {"Container": {"defaultRequest": {"cpu": "100m"}}}
        filename = "/tmp/{}.yaml".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            f.write(salt.utils.stringutils.to_str(request))

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.1)
        res = k8s.create_limit_range(
            "default",
            limits,
            name=name,
            apiserver_url="http://127.0.0.1:8080",
            update=True,
        )
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "limits", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        limit = (
            kubectl_out.get("spec").get("limits")[0].get("defaultRequest").get("cpu")
        )
        self.assertEqual("100m", limit)

    def test_get_limit_ranges(self):
        res = k8s.get_limit_ranges("default", apiserver_url="http://127.0.0.1:8080")
        a = len(res.get("items", []))
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "limits", "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = len(kubectl_out.get("items", []))
        self.assertEqual(a, b)

    def test_get_one_limit_range(self):
        name = self.name
        request = """
apiVersion: v1
kind: LimitRange
metadata:
  name: {}
spec:
  limits:
  - default:
      cpu: 200m
      memory: 512Mi
    defaultRequest:
      cpu: 100m
      memory: 256Mi
    type: Container
""".format(
            name
        )
        filename = "/tmp/{}.yaml".format(name)
        with salt.utils.files.fopen(filename, "w") as f:
            f.write(salt.utils.stringutils.to_str(request))

        create = Popen(
            ["kubectl", "--namespace=default", "create", "-f", filename], stdout=PIPE
        )
        # wee need to give kubernetes time save data in etcd
        time.sleep(0.1)
        res = k8s.get_limit_ranges(
            "default", name, apiserver_url="http://127.0.0.1:8080"
        )
        a = res.get("metadata", {}).get("name", "a")
        proc = Popen(
            ["kubectl", "--namespace=default", "get", "limits", name, "-o", "json"],
            stdout=PIPE,
        )
        kubectl_out = salt.utils.json.loads(proc.communicate()[0])
        b = kubectl_out.get("metadata", {}).get("name", "b")
        self.assertEqual(a, b)



"""
Support for nginx
"""

import re
import urllib.request

import salt.utils.decorators as decorators
import salt.utils.path


# Cache the output of running which('nginx') so this module
# doesn't needlessly walk $PATH looking for the same binary
# for nginx over and over and over for each function herein
@decorators.memoize
def __detect_os():
    return salt.utils.path.which("nginx")


def __virtual__():
    """
    Only load the module if nginx is installed
    """
    if __detect_os():
        return True
    return (
        False,
        "The nginx execution module cannot be loaded: nginx is not installed.",
    )


def version():
    """
    Return server version from nginx -v

    CLI Example:

    .. code-block:: bash

        salt '*' nginx.version
    """
    cmd = "{} -v".format(__detect_os())
    out = __salt__["cmd.run"](cmd).splitlines()
    ret = out[0].rsplit("/", maxsplit=1)
    return ret[-1]


def build_info():
    """
    Return server and build arguments

    CLI Example:

    .. code-block:: bash

        salt '*' nginx.build_info
    """
    ret = {"info": []}
    out = __salt__["cmd.run"]("{} -V".format(__detect_os()))

    for i in out.splitlines():
        if i.startswith("configure argument"):
            ret["build arguments"] = re.findall(r"(?:[^\s]*'.*')|(?:[^\s]+)", i)[2:]
            continue

        ret["info"].append(i)

    return ret


def configtest():
    """
    test configuration and exit

    CLI Example:

    .. code-block:: bash

        salt '*' nginx.configtest
    """
    ret = {}

    cmd = "{} -t".format(__detect_os())
    out = __salt__["cmd.run_all"](cmd)

    if out["retcode"] != 0:
        ret["comment"] = "Syntax Error"
        ret["stderr"] = out["stderr"]
        ret["result"] = False

        return ret

    ret["comment"] = "Syntax OK"
    ret["stdout"] = out["stderr"]
    ret["result"] = True

    return ret


def signal(signal=None):
    """
    Signals nginx to start, reload, reopen or stop.

    CLI Example:

    .. code-block:: bash

        salt '*' nginx.signal reload
    """
    valid_signals = ("start", "reopen", "stop", "quit", "reload")

    if signal not in valid_signals:
        return

    # Make sure you use the right arguments
    if signal == "start":
        arguments = ""
    else:
        arguments = " -s {}".format(signal)
    cmd = __detect_os() + arguments
    out = __salt__["cmd.run_all"](cmd)

    # A non-zero return code means fail
    if out["retcode"] and out["stderr"]:
        ret = out["stderr"].strip()
    # 'nginxctl configtest' returns 'Syntax OK' to stderr
    elif out["stderr"]:
        ret = out["stderr"].strip()
    elif out["stdout"]:
        ret = out["stdout"].strip()
    # No output for something like: nginxctl graceful
    else:
        ret = 'Command: "{}" completed successfully!'.format(cmd)
    return ret


def status(url="http://127.0.0.1/status"):
    """
    Return the data from an Nginx status page as a dictionary.
    http://wiki.nginx.org/HttpStubStatusModule

    url
        The URL of the status page. Defaults to 'http://127.0.0.1/status'

    CLI Example:

    .. code-block:: bash

        salt '*' nginx.status
    """
    resp = urllib.request.urlopen(url)
    status_data = resp.read()
    resp.close()

    lines = status_data.splitlines()
    if not len(lines) == 4:
        return
    # "Active connections: 1 "
    active_connections = lines[0].split()[2]
    # "server accepts handled requests"
    # "  12 12 9 "
    accepted, handled, requests = lines[2].split()
    # "Reading: 0 Writing: 1 Waiting: 0 "
    _, reading, _, writing, _, waiting = lines[3].split()
    return {
        "active connections": int(active_connections),
        "accepted": int(accepted),
        "handled": int(handled),
        "requests": int(requests),
        "reading": int(reading),
        "writing": int(writing),
        "waiting": int(waiting),
    }



#
# Author: Tyler Johnson <tjohnson@saltstack.com>
#

"""
Idem Support
============

This state provides access to idem states

.. versionadded:: 3002
"""
import pathlib
import re

__virtualname__ = "idem"


def __virtual__():
    if "idem.hub" in __utils__:
        return __virtualname__
    else:
        return False, "idem is not available"


def _get_refs(sources, tree):
    """
    Determine where the sls sources are
    """
    sls_sources = []
    SLSs = []
    if tree:
        sls_sources.append("file://{}".format(tree))
    for sls in sources:
        path = pathlib.Path(sls)
        if path.is_file():
            ref = str(path.stem if path.suffix == ".sls" else path.name)
            SLSs.append(ref)
            implied = "file://{}".format(path.parent)
            if implied not in sls_sources:
                sls_sources.append(implied)
        else:
            SLSs.append(sls)
    return sls_sources, SLSs


def _get_low_data(low_data):
    """
    Get salt-style low data from an idem state name
    """
    # state_|-id_|-name_|-function
    match = re.match(r"(.+)_\|-(.+)_\|-(.+)_\|-(.+)", low_data)
    return {
        "state": match.group(1),
        "__id__": match.group(2),
        "name": match.group(3),
        "fun": match.group(4),
    }


def state(
    name,
    sls,
    acct_file=None,
    acct_key=None,
    acct_profile=None,
    cache_dir=None,
    render=None,
    runtime=None,
    source_dir=None,
    test=False,
):
    """
    Execute an idem sls file through a salt state

    sls
        A list of idem sls files or sources

    acct_file
        Path to the acct file used in generating idem ctx parameters.
        Defaults to the value in the ACCT_FILE environment variable.

    acct_key
        Key used to decrypt the acct file.
        Defaults to the value in the ACCT_KEY environment variable.

    acct_profile
        Name of the profile to add to idem's ctx.acct parameter
        Defaults to the value in the ACCT_PROFILE environment variable.

    cache_dir
        The location to use for the cache directory

    render
        The render pipe to use, this allows for the language to be specified (jinja|yaml)

    runtime
        Select which execution runtime to use (serial|parallel)

    source_dir
        The directory containing sls files

    .. code-block:: yaml

        cheese:
            idem.state:
                - runtime: parallel
                - sls:
                    - idem_state.sls
                    - sls_source

    :maturity:      new
    :depends:       acct, pop, pop-config, idem
    :platform:      all
    """
    hub = __utils__["idem.hub"]()

    if isinstance(sls, str):
        sls = [sls]

    sls_sources, SLSs = _get_refs(sls, source_dir or hub.OPT.idem.tree)

    coro = hub.idem.state.apply(
        name=name,
        sls_sources=sls_sources,
        render=render or hub.OPT.idem.render,
        runtime=runtime or hub.OPT.idem.runtime,
        subs=["states"],
        cache_dir=cache_dir or hub.OPT.idem.cache_dir,
        sls=SLSs,
        test=test,
        acct_file=acct_file or hub.OPT.acct.acct_file,
        acct_key=acct_key or hub.OPT.acct.acct_key,
        acct_profile=acct_profile or hub.OPT.acct.acct_profile or "default",
    )
    hub.pop.Loop.run_until_complete(coro)

    errors = hub.idem.RUNS[name]["errors"]
    success = not errors

    running = []
    for idem_name, idem_return in hub.idem.RUNS[name]["running"].items():
        standardized_idem_return = {
            "name": idem_return["name"],
            "changes": idem_return["changes"],
            "result": idem_return["result"],
            "comment": idem_return.get("comment"),
            "low": _get_low_data(idem_name),
        }
        running.append(standardized_idem_return)

    return {
        "name": name,
        "result": success,
        "comment": "Ran {} idem states".format(len(running)) if success else errors,
        "changes": {},
        "sub_state_run": running,
    }



import pathlib

import pytest

from salt.wheel import file_roots


def _make_temp_root_file(root, *subpaths, binary=False, dir_only=False):
    """
    Creates a file under the specified subpaths of the given root with the filepath as its content.
    """
    full_path = pathlib.Path(root, *subpaths)
    full_path.parent.mkdir(exist_ok=True, parents=True)
    if not dir_only:
        if binary:
            content = b"\x00"
            full_path.write_bytes(content)
        else:
            content = str(full_path)
            full_path.write_text(content)


@pytest.fixture
def base_root_1(tmp_path):
    path = tmp_path / "base_root_1"
    path.mkdir()
    return path


@pytest.fixture
def base_root_2(tmp_path):
    path = tmp_path / "base_root_2"
    path.mkdir()
    return path


@pytest.fixture
def prod_root_1(tmp_path):
    path = tmp_path / "prod_root_1"
    path.mkdir()
    return path


@pytest.fixture
def prod_root_2(tmp_path):
    path = tmp_path / "prod_root_2"
    path.mkdir()
    return path


@pytest.fixture
def populated_roots(
    base_root_1,
    base_root_2,
    prod_root_1,
    prod_root_2,
):
    roots = {
        "base": [str(base_root_1), str(base_root_2)],
        "prod": [str(prod_root_1), str(prod_root_2)],
    }

    _make_temp_root_file(base_root_1, "test_base_1_file")
    _make_temp_root_file(base_root_1, "common_file")
    _make_temp_root_file(base_root_1, "base_1_subdir", "test_base_1_file_in_subdir")
    _make_temp_root_file(base_root_1, "base_1_subdir", "common_file")
    _make_temp_root_file(
        base_root_1, "base_1_subdir", "test_base_1_file_in_subdir_binary", binary=True
    )
    _make_temp_root_file(
        base_root_1,
        "base_1_subdir",
        "base_1_sub_subdir",
        "test_base_1_file_in_sub_subdir",
    )
    _make_temp_root_file(base_root_2, "test_base_2_file_1")
    _make_temp_root_file(base_root_2, "test_base_2_file_2")
    _make_temp_root_file(base_root_2, "common_file")
    _make_temp_root_file(prod_root_1, "test_prod_2_file")
    _make_temp_root_file(prod_root_1, "common_file")
    _make_temp_root_file(prod_root_1, "prod_1_subdir", dir_only=True)

    return roots


@pytest.fixture
def base_list_env(
    base_root_1,
    base_root_2,
):
    return {
        str(base_root_1): {
            "base_1_subdir": {
                "base_1_sub_subdir": {"test_base_1_file_in_sub_subdir": "f"},
                "test_base_1_file_in_subdir_binary": "f",
                "common_file": "f",
                "test_base_1_file_in_subdir": "f",
            },
            "common_file": "f",
            "test_base_1_file": "f",
        },
        str(base_root_2): {
            "common_file": "f",
            "test_base_2_file_2": "f",
            "test_base_2_file_1": "f",
        },
    }


@pytest.fixture
def prod_list_env(
    prod_root_1,
    prod_root_2,
):
    return {
        str(prod_root_1): {"common_file": "f", "test_prod_2_file": "f"},
        str(prod_root_2): {},
    }


@pytest.fixture
def configure_loader_modules(populated_roots):
    return {
        file_roots: {
            "__opts__": {"file_roots": populated_roots},
        },
    }


def test_find(base_root_1, base_root_2):
    file_name = "common_file"
    expected = [
        {str(base_root_1 / file_name): "txt"},
        {str(base_root_2 / file_name): "txt"},
    ]
    ret = file_roots.find(file_name)
    assert ret == expected


def test_find_prod(prod_root_1):
    file_name = "common_file"
    expected = [{str(prod_root_1 / file_name): "txt"}]
    ret = file_roots.find(file_name, saltenv="prod")
    assert ret == expected


def test_find_in_subdir(base_root_1):
    file_name = pathlib.Path("base_1_subdir", "test_base_1_file_in_subdir")
    expected = [{str(base_root_1 / file_name): "txt"}]
    ret = file_roots.find(str(file_name))
    assert ret == expected


def test_find_does_not_exist():
    file_name = "prod_1_subdir"
    expected = []
    ret = file_roots.find(str(file_name), saltenv="prod")
    assert ret == expected


def test_find_binary(base_root_1):
    file_name = pathlib.Path("base_1_subdir", "test_base_1_file_in_subdir_binary")
    expected = [{str(base_root_1 / file_name): "bin"}]
    ret = file_roots.find(str(file_name))
    assert ret == expected


def test_list_env(base_list_env):
    ret = file_roots.list_env()
    assert ret == base_list_env


def test_list_env_prod(prod_list_env):
    ret = file_roots.list_env(saltenv="prod")
    assert ret == prod_list_env


def test_list_roots(base_list_env, prod_list_env):
    expected = {"base": [base_list_env], "prod": [prod_list_env]}
    ret = file_roots.list_roots()
    assert ret == expected


def test_read(base_root_1, base_root_2):
    file_name = "common_file"
    root_1_file = str(base_root_1 / file_name)
    root_2_file = str(base_root_2 / file_name)
    expected = [{root_1_file: root_1_file}, {root_2_file: root_2_file}]
    ret = file_roots.read(file_name)
    assert ret == expected


def test_read_prod(prod_root_1):
    file_name = "common_file"
    root_1_file = str(prod_root_1 / file_name)
    expected = [{root_1_file: root_1_file}]
    ret = file_roots.read(file_name, saltenv="prod")
    assert ret == expected


def test_read_binary():
    file_name = pathlib.Path("base_1_subdir", "test_base_1_file_in_subdir_binary")
    ret = file_roots.read(str(file_name))
    assert ret == []


def test_read_in_subdir(base_root_1):
    file_name = pathlib.Path("base_1_subdir", "test_base_1_file_in_subdir")
    subdir_file = str(base_root_1 / file_name)
    expected = [{subdir_file: subdir_file}]
    ret = file_roots.read(str(file_name))
    assert ret == expected


def test_write(base_root_1):
    file_name = "testfile"
    ret = file_roots.write(file_name, file_name)
    assert f"Wrote data to file {str(base_root_1 / file_name)}" in ret


def test_write_index(base_root_2):
    file_name = "testfile"
    ret = file_roots.write(file_name, file_name, index=1)
    assert f"Wrote data to file {str(base_root_2 / file_name)}" in ret


def test_write_prod(prod_root_2):
    file_name = "testfile"
    ret = file_roots.write(file_name, file_name, saltenv="prod", index=1)
    assert f"Wrote data to file {str(prod_root_2 / file_name)}" in ret


def test_write_subdir(prod_root_1):
    file_name = str(pathlib.Path("prod_1_subdir", "testfile"))
    ret = file_roots.write(file_name, file_name, saltenv="prod")
    assert f"Wrote data to file {str(prod_root_1 / file_name)}" in ret


def test_write_make_new_subdir(prod_root_2):
    file_name = str(pathlib.Path("prod_2_subdir", "testfile"))
    ret = file_roots.write(file_name, file_name, saltenv="prod", index=1)
    assert f"Wrote data to file {str(prod_root_2 / file_name)}" in ret


def test_write_invalid_env():
    file_name = "testfile"
    env = "not_an_env"
    ret = file_roots.write(file_name, file_name, saltenv=env)
    assert f"{env} is not present" in ret


def test_write_invalid_index():
    file_name = "testfile"
    ret = file_roots.write(file_name, file_name, index=2)
    assert "index 2 in environment base is not present" in ret


def test_write_invalid_absolute_path(base_root_1):
    file_name = str(base_root_1 / "testfile")
    ret = file_roots.write(file_name, file_name)
    assert "is not relative to the environment" in ret


def test_write_invalid_path():
    file_name = str(pathlib.Path("..", "testfile"))
    ret = file_roots.write(file_name, file_name)
    assert "Invalid path: " in ret



# Python libs

import pytest

# Salt libs
import salt.beacons.salt_proxy as salt_proxy
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {salt_proxy: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = salt_proxy.validate(config)

    assert ret == (False, "Configuration for salt_proxy beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = salt_proxy.validate(config)

    assert ret == (False, "Configuration for salt_proxy beacon requires proxies.")


def test_salt_proxy_running():
    mock = MagicMock(return_value={"result": True})
    with patch.dict(salt_proxy.__salt__, {"salt_proxy.is_running": mock}):
        config = [{"proxies": {"p8000": ""}}]

        ret = salt_proxy.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = salt_proxy.beacon(config)
        assert ret == [{"p8000": "Proxy p8000 is already running"}]


def test_salt_proxy_not_running():
    is_running_mock = MagicMock(return_value={"result": False})
    configure_mock = MagicMock(
        return_value={
            "result": True,
            "changes": {
                "new": "Salt Proxy: Started proxy process for p8000",
                "old": [],
            },
        }
    )
    cmd_run_mock = MagicMock(
        return_value={"pid": 1000, "retcode": 0, "stderr": "", "stdout": ""}
    )
    with patch.dict(
        salt_proxy.__salt__, {"salt_proxy.is_running": is_running_mock}
    ), patch.dict(
        salt_proxy.__salt__, {"salt_proxy.configure_proxy": configure_mock}
    ), patch.dict(
        salt_proxy.__salt__, {"cmd.run_all": cmd_run_mock}
    ):
        config = [{"proxies": {"p8000": ""}}]

        ret = salt_proxy.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = salt_proxy.beacon(config)
        assert ret == [{"p8000": "Proxy p8000 was started"}]
